# encoding: UTF-8

# Copyright 2012, by Mark Jason Grundman
#
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# -----------------------------------------------------------------------------
#
#	Name        :   SPGears_mjg.rb
#	Description :   Draws various types of gears with involute tooth profiles
#	Menu Item   :   Plugins (Extensions) -> SP Gears (MJG)
#	Context Menu:   NONE
#	Date        :   2012-06-28
#	Version     :   1.3.0
#
#	Doug Herrrman's gear plugin ('gear.3.rb' created in 2006) does a great job
# of creating a 2D gear profile.  Unfortunately when used to create gears for
# SketchyPhysics (SP), it did not create accurate collision geometry.
# Doug's plug-in inspired (and provided a basis for) this re-work of the
# concept, which solves the collision geometry problems when used with SP.
#
#	This is a great utility for creating gears in 'static' drawings, too.
# I tried to cover as many common gear styles and variations as possible.
#
#	v1.3 : 2021-10-26  by Dan Rathbun
#      - Properly namespace wrapped and packaged as a RBZ extension archive.
#      - General cleanup of indents, extra long lines and empty space lines.
#      - Corrected inputbox validation loop in SPGear::dialog().
#        * Was calling the same method recursively and not restoring the
#            users previous inputs.
#        * Was using bad index to validate Bevel Angle results which did
#            not cause the inputbox to redisplay allowing error in later code
#            when the angle was out of range.
#
#	v1.2 : 2012-06-28
#      - revisions and additions
#			 - added root fillets
#			 - added rim and web to base profile
#			 - modified involute calculations to maintain curve detail as size decreases
#			
#	v1.1 - minor bug fix
#
#	v1.0 - initial release
#
#-----------------------------------------------------------------------------/

module MJGrundman
  module SPGears

    # Generates an Involute Tooth Gear.
    class SPGear

      begin
        WIN = Sketchup.platform == :platform_win
      rescue NoMethodError
        WIN = !(RUBY_PLATFORM =~ /darwin/i) # NO !~ method for Ruby < 2.x
      end

      # default values for dialog box
      @@teeth = 15
      @@pressure_angle = 20.0  # degrees
      @@rad_pitch = 5.0.inch
      @@face_width = 3.0
      @@helix_angle = 0.0
      @@bevel_angle = 0.0
      @@shaft_diameter = 1.0

      def SPGear.dialog
        # First prompt for the dimensions.
        prompts = [
          "Number of Teeth: ", "Pressure Angle: ", "Pitch Radius/Rack Leng.: ",
          "Face Width: ", "Helix Angle: ", "Bevel Angle: ", "Shaft Diameter: "
        ]
        values = [
          @@teeth, @@pressure_angle, @@rad_pitch, @@face_width,
          @@helix_angle, @@bevel_angle, @@shaft_diameter
        ]

        begin
          # Display the inputbox
          results = inputbox(
            prompts, values,
            [nil,'14.5|20.0|25.0',nil,nil,nil, nil,nil],
            "Gear Parameters"
          )
          return if !results # User cancelled the operation

          # Validate Bevel Angle input:
          ba_index = prompts.index("Bevel Angle: ")
          bevel_angle = results[ba_index].abs
          if bevel_angle < 0.0 || bevel_angle > 180.0
            UI.messagebox(
              'Bevel Angle must be between 0 and (+)180 degrees.'<<
              "\n\nPlease try again."
            )
            values = results # So the user's inputs are redisplayed !
            fail(RangeError, 'Bevel Input Error',[])
          else
            results[ba_index] = results[ba_index].abs
          end
        rescue RangeError => err
          (err.message == 'Bevel Input Error') ? retry : (return false)
        rescue => err
          UI.messagebox(err.message)
          return false
        end

        # update default settings
        @@teeth, @@pressure_angle, @@rad_pitch, @@face_width,
        @@helix_angle, @@bevel_angle, @@shaft_diameter = results

        begin
          SPGear.new
        rescue => err          
          puts err.inspect
          puts err.backtrace
          if WIN
            drive, path, line, method = err.backtrace[0].split(':')
            path = drive.concat(':',path)
          else
            path, line, method = err.backtrace[0].split(':')
          end
          msg = EXTENSION.name.dup
          msg << "\n\n#{err.class}: #{err.message}\n"
          msg << "#{method.gsub('in `','In method: "').gsub("'",'"')}\n"
          msg << "At line: #{line}\n"
          msg << "File: \"#{File.join(path.split('/')[-2..-1])}\""
          UI.messagebox(msg)
        end

      end ### class method dialog()

      def initialize(
        teeth = @@teeth,
        pressure_angle = @@pressure_angle,
        rad_pitch = @@rad_pitch,
        face_width = @@face_width,
        helix_angle = @@helix_angle,
        bevel_angle = @@bevel_angle,
        shaft_diameter = @@shaft_diameter
      )

        @teeth = teeth.abs
        @pressure_angle = pressure_angle.degrees
        @rad_pitch = rad_pitch.to_l
        @face_width = face_width.to_l
        @helix_angle = helix_angle.degrees
        @bevel_angle = bevel_angle.degrees
        @shaft_diameter = shaft_diameter.to_l
        @rad_bore = shaft_diameter / 2

        # Check face width of bevel gears and reduce to prevent inversion
        # at cone apex of bevel gears.
        if bevel_angle > 0 && bevel_angle < 180
          maxFaceWidth = @rad_pitch / Math.sin(@bevel_angle)
          if @face_width > maxFaceWidth
            @face_width = maxFaceWidth - 0.1
            paramChange = "Face width parameter exceeded maximum\n" <<
              "  for bevel angle and pitch radius parameters.\n\n" <<
              "Face width was reduced to fit."
            UI.messagebox(paramChange)
          end
        end

        # Calculate various gear attributes, from user input, for later
        # calculations and other uses.
        @pitch_diameter = 2.0 * @rad_pitch
        @diametral = @teeth / @pitch_diameter			# number of teeth/inch	   
        @addendum = 1 / @diametral
        if @diametral < 20
          @whole_depth = 2.157 / @diametral
        else
          @whole_depth = 2.2 / @diametral + 0.002
        end
        @rad_outside = ((@teeth + 2) / @diametral) / 2		# measured at tooth tip (top land)
        @rad_inside = @rad_outside - (@addendum * 2)		# outside radius, less the working depth
        @rad_root = @rad_outside - @whole_depth
        @rad_base = @rad_pitch * Math.cos(@pressure_angle)
        @rad_fillet = @rad_inside - @rad_root
        @circular_pitch = 180.degrees / @diametral

        # select the number of arms for the hub
        case @pitch_diameter.abs
          when 0.mm..299.mm then @arms = 3.0
          when 300.mm..499.mm then @arms = 5.0
          when 500.mm..1499.mm then @arms = 6.0
          when 1500.mm..2399.mm then @arms = 8.0
          else @arms = 10.0
        end

        #~ @wdth_armHub = @diametral * 9.5
        #~ @wdth_armRim = @diametral * 7.8
        @rad_hub = @shaft_diameter * 1.4
        @len_hub = @shaft_diameter * 1.5

        # Encapsulate into a single UNDO operation.
        model = Sketchup.active_model
        begin
          model.start_operation("Create Gear(mjg)",true)
          ###
            # add a group to contain output geometry
            entities = model.active_entities
            @mainGroup = entities.add_group
            spDrawGear()
            #
          ###
          model.commit_operation
        rescue
          # Operations left open now spit out a warning !
          model.abort_operation
          puts 'Model operation aborted ...'
          raise # the exception back to the dialog() class method
        end

      end ### SPGear class initializer

      #==========================================================================

      #  -- NOTE: The SkectchyPhysics plugin/extension overrides the SU API
      #  --  'group.copy' method.
      #  -- Using components as work-around greatly increases drawing speed and
      #  --  enables'copying' groups without interference from SketchyPhysics.

      private

      # Method generates a wedge/slice that includes a sinlge gear tooth,
      # tooth space, and attached rim/hub/base, then copies are rotated around
      # the pitch circle to create a complete gear.
      def spDrawGear
        toothPts = []
        basePts = []

        # Determine where involute intersects inside radius, then add fillet
        # points to the base profile (call spFilletPoints)
        if @rad_inside <= @rad_base
          basePts = spFilletPoints(@rad_inside, -@rad_fillet)
          ang = 0
        else
          s = Math.sqrt(@rad_inside * @rad_inside - @rad_base * @rad_base)
          ang = s / @rad_base
          x0, y0 = spGearInvolute(ang)
          basePts = spFilletPoints(x0, y0 - @rad_fillet)
        end

        # Determine where Involute curve intersects outside (tip) radius
        s_max = Math.sqrt(@rad_outside * @rad_outside - @rad_base * @rad_base)
        ang_max = s_max / @rad_base

        # Calculate angle variable for involute curve
        k = 135.degrees
        angNew = (k / @teeth)

        # Set starting point and length of involute curve
        toothPts.push(basePts.last)
        ang += angNew

        # Create points for the tooth profile along tooth's right side
        # (call spGearInvolute).
        while ang < ang_max
          x, y = spGearInvolute(ang)
          if x < @rad_inside
            toothPts[0] = [x, y, 0]
            ang += angNew
          else
            toothPts.push([x, y, 0])
            ang += angNew
          end
        end

        # Create final point of involute where the curve intersects the
        # outside (tip) diameter of gear.
        x, y = spGearInvolute(ang_max)
        toothPts.push([x, y, 0])

        # Create points for the rim and add to the base profile
        vect = toothPts[0].vector_to([0,0,0])
        vect.length = @whole_depth * 0.6 + @rad_fillet
        t = Geom::Transformation.translation vect
        basePts.push(toothPts[0].transform(t))

        # Copy base profile points into the tooth profile.
        toothPts = basePts + toothPts

        # Rotate profile so that involute intersects the x-axis at the
        # pitch radius.
        s = Math.sqrt(@rad_pitch**2 - @rad_base**2)
        delta = (s / @rad_base) - @pressure_angle
        t0 = Geom::Transformation.rotation [0,0,0], Z_AXIS, -delta
        toothPts.collect!{|pnt| t0 * pnt}

        # Reflect tooth profile about the x-axis.
        t1 = Geom::Transformation.rotation [0,0,0], X_AXIS, 180.degrees
        reflectPts = toothPts.collect{|pnt| t1 * pnt}

        # Rotate relected points by one tooth space.
        t2 = Geom::Transformation.rotation [0,0,0], Z_AXIS, 180.degrees/@teeth
        reflectPts.collect!{|pnt| t2 * pnt}

        # Reverse the order of reflected points and add to the tooth profile.
        profilePts = toothPts + reflectPts.reverse!

        # Rotate tooth profile to center along the X-axis and save profile
        # points as a group.
        t3 = Geom::Transformation.rotation [0,0,0], Z_AXIS, -180.degrees/(2*@teeth)
        profilePts.collect!{|pnt| t3 * pnt }

        # Make sure all points are set on the same plane.
        profilePts.each{|pnt| pnt.z = 0 if pnt.z.between?(-0.000001, 0.000001)}

        # Extrude the profile into a complete tooth (call spExtrudeProfile).
        grpProfile = spExtrudeProfile(profilePts)

      end  # spDrawGear	

      # Generate a mesh for each 'face' of the tooth or base by copying
      # original profile as tarnsformation cross-sections.
      def spExtrudeProfile(pointsArray)
        subGroup = @mainGroup.entities.add_group
        ent = subGroup.entities

        # for beveled gears, mesh is rotated to bevel angle, during extrusion, then rotated back to vertical
        t0 = Geom::Transformation.rotation [@rad_pitch,0,0], Y_AXIS, -@bevel_angle
        t1 = Geom::Transformation.rotation [@rad_pitch,0,0], Y_AXIS, @bevel_angle

        #  add original profile as starting cross section (end-face) to the extrusion mesh
        pts = pointsArray.collect{|pnt| t0 * pnt}
        profilePts = []
        profilePts.push(pts)

        # count at leaast one cross-section for closing end-face
        sectionCnt = 1.0

        # count additional cross-sections for helical/spiral and bevel to reduce twisting/stretching distortion
        UI.messagebox("Diametral = " << @diametral.to_s)
        if @helix_angle != 0
          sectionCnt += (@helix_angle.abs/7.0).floor
          sectionCnt += (@face_width.abs/(2*@diametral)).floor
          sectionCnt += (Math.tan(@bevel_angle) * 2).floor

          # set helical rotation angle (per section)
          arcLength = Math.tan(@helix_angle.degrees) * @face_width
          centerAngle = arcLength / @rad_pitch
          arcSection = centerAngle / sectionCnt
        end

        # set seperation width (along z-axis) per cross-section
        sectionWidth = @face_width / sectionCnt

        # calculate and apply cross-section transformations 
        1.upto(sectionCnt) do |sect|

          # set default transformation for spur gear 
          z0 = sectionWidth * sect
          t3 = Geom::Transformation.new [0,0,z0]

          # for beveled gears, scale radius and secction width along tapered facewidth
          if @bevel_angle > 0 && @bevel_angle < 180.degrees

            #calculate scaling factor for radius (slope run on x-axis)
            maxConeDistance = @rad_pitch / Math.sin(@bevel_angle)
            s0 = 1 - ((sectionWidth * sect) / maxConeDistance)
            s0 = 0.002 if s0 <= 0.001		# limit the minimum scaling factor

            # calculate new section width based on scaling factor (slope rise on z-axis)
            x0 = @rad_pitch - (@rad_pitch * s0)
            z0 = x0 / Math.sin(@bevel_angle)
            t3 = Geom::Transformation.translation [0,0,z0]
            t4 = Geom::Transformation.scaling [@rad_pitch,0,0], s0, s0, 1

          end # end if

          # add rotations for helical/spiral
          t5 = Geom::Transformation.rotation [0,0,0], Z_AXIS, arcSection * sect if @helix_angle != 0

          # apply combined transformations
          if @bevel_angle > 0 && @helix_angle != 0
            tr = t5 * t0 * t4 * t3
          elsif @bevel_angle > 0
            tr = t0 * t4 * t3
          elsif @helix_angle != 0
            tr = t5 * t3
          else
            tr = t3
          end

          #add transformed points to extusion mesh then loop to the next section
          pts = pointsArray.collect{|pnt| tr * pnt}
          profilePts.push(pts)

        end # cross-secction transformations

        # separate extruded mesh points into subgroups
        rimPts = []
        toothPts = []
        profilePts.each{|sect|
          toothPts.push(sect[5..-6])

          pts = sect[0..5], sect[-6..-1]
          rimPts.push(pts[0] + pts[1])
        }

        # finish by creating extrusion faces for each subgroup (call function specific to each group)
        grpTooth = spDrawTooth(toothPts, subGroup)
        grpRim, basePts = spDrawRim(rimPts, subGroup)
        grpBase = spDrawBase(basePts, grpRim)

      end # spExtrudeProfile

      def spDrawTooth(pointsArray, group)
        sectionMax = pointsArray.size - 2
        pointMax = pointsArray[0].size - 2
        grpTooth = group.entities.add_group

         # draw tooth surface faces
        land1 = (pointsArray[0].size / 2) -2
        land2 = land1 + 1

        0.upto(sectionMax){|r|
          0.upto(pointMax){|c|
            faces = []
            faceA = grpTooth.entities.add_face(pointsArray[r][c+1], pointsArray[r+1][c], pointsArray[r][c])
            (faceA.edges[2].soft="true"; faceA.edges[2].smooth="true"; ) unless r == 0
            (faceA.edges[1].soft="true"; faceA.edges[1].smooth="true"; ) if c == 0
            faceB = grpTooth.entities.add_face(pointsArray[r+1][c+1], pointsArray[r][c+1], pointsArray[r+1][c])
            (faceB.edges[0].soft="true"; faceB.edges[0].smooth="true"; ) unless (c == land1 || c == land2)
            (faceB.edges[2].soft="true"; faceB.edges[2].smooth="true"; )
          }
        }

        # draw tooth end-faces
        [pointsArray.first, pointsArray.last].each do |pts|
          face = grpTooth.entities.add_face(pts)
          face.edges.each{|edge|
            if edge.end.position == pts.first
              edge.soft="true"; edge.smooth="true"
            end
          }
        end # do

        newToothGroup = spCopyGroup(grpTooth, "convexhull", false)

      end # spDrawTooth

      def spDrawRim(pointsArray, group)
        sectionMax = pointsArray.size - 2
        pointMax = pointsArray[0].size - 2
        grpRim = group.entities.add_group

        # extend rim points to include a tooth space
        tr = Geom::Transformation.rotation [0,0,0], Z_AXIS, (360.degrees / @teeth)
        pointsArray.each{|sect|

          # remove and relocate fillet points
          pts = sect.slice!(0..3)
          pts.collect!{|pnt| tr * pnt}
          pts.each{|pnt| sect.push(pnt)}

          # copy rim point, leaving orginial in place
          pnt = sect.at(0)
          sect.push(pnt.transform(tr))
        }

        sectionMax = pointsArray.size - 2
        pointMax = pointsArray[0].size - 2
        grpRim = group.entities.add_group

        # draw rim end-faces
        [pointsArray.first, pointsArray.last].each do |pts|

          faceA = grpRim.entities.add_face(pts.values_at(0..3))
          last = (pts.size) -1
          faceB1 = grpRim.entities.add_face(pts.values_at(3..7))
          faceB2 = grpRim.entities.add_face(pts.values_at(3,7,8,last))
          faceB3 = grpRim.entities.add_face(pts.values_at(8..last))
          faces = faceA, faceB1, faceB2, faceB3
          faces.each{|face| face.reverse! if @bevel_angle > 0 && pts == pointsArray.first}

          (faceA.edges[0].soft="true"; faceA.edges[0].smooth="true"; )
          (faceA.edges[1].soft="true"; faceA.edges[1].smooth="true"; )# if pts == pointsArray.last
          (faceA.edges[2].soft="true"; faceA.edges[2].smooth="true"; )
          (faceA.edges[3].soft="true"; faceA.edges[3].smooth="true"; )# if pts == pointsArray.first

      #~ #0 		(faceB1.edges[0].soft="true"; faceB1.edges[0].smooth="true"; ) if pts == pointsArray.last
      #~ #0		(faceB1.edges[1].soft="true"; faceB1.edges[1].smooth="true"; ) if pts == pointsArray.first
      #~ #0		(faceB1.edges[2].soft="true"; faceB1.edges[2].smooth="true"; )
      #~ #0		(faceB1.edges[3].soft="true"; faceB1.edges[3].smooth="true"; )
      #~ #0		(faceB1.edges[4].soft="true"; faceB1.edges[4].smooth="true"; ) if pts == pointsArray.last

          (faceB2.edges[0].soft="true"; faceB2.edges[0].smooth="true"; )
          (faceB2.edges[1].soft="true"; faceB2.edges[1].smooth="true"; ) if pts != pointsArray.last #&& pts != pointsArray.first
          (faceB2.edges[2].soft="true"; faceB2.edges[2].smooth="true"; )
          (faceB2.edges[3].soft="true"; faceB2.edges[3].smooth="true"; ) if pts != pointsArray.first #&& pts != pointsArray.first

      #~ #0		(faceB3.edges[0].soft="true"; faceB3.edges[0].smooth="true"; )
      #~ #0		(faceB3.edges[1].soft="true"; faceB3.edges[1].smooth="true"; )
          (faceB3.edges[2].soft="true"; faceB3.edges[2].smooth="true"; ) if pts == pointsArray.first
          (faceB3.edges[3].soft="true"; faceB3.edges[3].smooth="true"; ) if pts == pointsArray.last
      #~ #0		(faceB3.edges[4].soft="true"; faceB3.edges[4].smooth="true"; )

        end # loop

        rimPts = []
        pointsArray.each{|sect|
          pts = sect.values_at(0, 3, -1, 4..-2)
          rimPts.push(pts)
        }

        # draw inner rim faces
        range = 3..(rimPts[0].size-2)
        index = range.to_a
        0.upto(sectionMax){|r|
          index.each do |c|
            c1 = c+1; r1 = r+1;
            faceA = grpRim.entities.add_face(rimPts[r][c], rimPts[r][c1], rimPts[r1][c1])
              (faceA.edges[0].soft="true"; faceA.edges[0].smooth="true"; ) unless r == 0
              (faceA.edges[1].soft="true"; faceA.edges[1].smooth="true"; )
              (faceA.edges[2].soft="true"; faceA.edges[2].smooth="true"; )
            faceB = grpRim.entities.add_face(rimPts[r1][c1], rimPts[r1][c], rimPts[r][c])
              (faceB.edges[0].soft="true"; faceB.edges[0].smooth="true"; ) unless r == sectionMax
              (faceB.edges[1].soft="true"; faceB.edges[1].smooth="true"; )
              (faceB.edges[2].soft="true"; faceB.edges[2].smooth="true"; )

          end # loop c
        } # loop r

        newRimGroup = spCopyGroup(grpRim, "convexhull", true)

        basePts = []
        rimPts.each{|sect| basePts.push(sect.values_at(0..2))}

        return newRimGroup, basePts

      end # spDrawRim

      def spDrawBase(pointsArray, group)
        sectionMax = pointsArray.size - 2
        pointMax = pointsArray[0].size - 2
        grpBase = group.entities.add_group
        entBase = grpBase.entities

        # calculate center point, depths, and radii for base and web
        refPnt = [pointsArray.first.first, pointsArray.last.first]
        refVector = [[0,0,refPnt[0].z].vector_to(refPnt[0]), [0,0,refPnt[1].z].vector_to(refPnt[1])]
        refRadius = [(refVector[0].length * 0.98), (refVector[1].length * 0.98)]
              #~ puts "refPnt: #{refPnt.inspect}"
              #~ puts "refVector: #{refVector.inspect}"
              #~ puts "refRadius: #{refRadius.inspect}"
        web_width = @circular_pitch / 2
        base_width = refPnt[1].z - refPnt[0].z
        web_depth = (base_width - web_width) / 2
        if base_width < (web_depth * 2 + web_width)
          web_width = base_width
          web_depth = 0
        end
              #~ puts "web_width: #{web_width}"
              #~ puts "base_width: #{base_width}"
              #~ puts "web_depth: #{web_depth}"
        if @bevel_angle > 0
          x0 = refRadius[0] - refRadius[1]
          y0 = refPnt[1].z - refPnt[0].z
          ang = Math.atan(y0/x0)
          rad0 = refRadius[0] - (web_depth * Math.cos(ang))
          rad1 = refRadius[1] + (web_depth * Math.cos(ang))
          web_radius = [rad0, rad1]			

        end # calculating variables

        # draw base end-faces
        basePts = []
        baseFaces = []
        [pointsArray.first, pointsArray.last].each do |sect|
          basePts = sect.values_at(0, 1)
          pts = []
          1.upto(@teeth){|n|
            tr = Geom::Transformation.rotation [0,0,0], Z_AXIS, (360.degrees * n / @teeth)
            basePts.each{|pnt| pts.push(pnt.transform(tr))}
          }
          basePts.push(pts)
          face = entBase.add_face(pts)
          face.reverse! if sect == pointsArray.first && @bevel_angle > 0
          face.edges.each{|edge| edge.soft="true"; edge.smooth="true";} if web_depth > 0
          baseFaces.push(face)

        end # draw base end-faces

        # draw the inner rim on the base faces, then pushpull them to the web depth		
        if web_depth > 0
          [0, 1].each{|n|
            centerPnt = [0, 0, refPnt[n].z]
            faceNorm = baseFaces[n].normal
            circle = entBase.add_circle(centerPnt, faceNorm, refRadius[n], 32.0)
            newFace = entBase[(entBase.count - 1)]
            newFace = newFace.pushpull -web_depth, false

            # locate the pushpull'ed face, which is replaced during the pushpull operation
              faces = circle[0].faces
              attachedFace = nil
            # from the initial circle defined above, find an adjacent, non-horizontal face (cylinder wall)
              faces.each{|f| attachedFace = f if f.normal != [0,0,1] && f.normal != [0,0,-1]}
              edges = attachedFace.edges
              edges.each{|edge|
            # excluding the previous (top) edge, find the adjacent horizontal face (cylinder floor)	
                unless edge == circle[0]
                  faces = []
                  edge.faces.each{|f| attachedFace = f if f.normal == [0,0,1] || f.normal == [0,0,-1]}
                end
              }

            # for bevel gears, scale the lower (wider) cylinder floor (edges) to taper the inner walls
            if @bevel_angle > 0 && n == 0
              if n == 0
                depth = web_depth
                s0 = web_radius[n] / refRadius[n]
                t0 = Geom::Transformation.scaling([0,0,depth], s0, s0, 1.0)
                entBase.transform_entities(t0, attachedFace.edges)
              end
            end
          }
        end # draw the inner rim

        # draw hub and shaft bore
        if @rad_bore > 0 && @len_hub > web_width

          #~ # make a hole for the hub
          #~ z0 = refPnt[0].z + web_depth
          #~ circle1 = entBase.add_circle([0,0,z0], Z_AXIS, @rad_hub, 32)
          #~ ents = entBase.to_a
          #~ face = ents.last 
          #~ face = entBase.add_face(circle1)
          #~ face.pushpull(-web_width, false)			

          #~ # insert hub and bore - v.1
          #~ z1 = (base_width - @len_hub) / 2
          #~ circle1 = entBase.add_circle([0,0,z1], Z_AXIS, @rad_hub, 32)
          #~ hub = entBase.add_face(circle1)
          #~ hub.pushpull(@len_hub, false)
          #~ circle2 = entBase.add_circle([0,0,z1], Z_AXIS, @rad_bore, 32)
          #~ entities = entBase.to_a
          #~ bore = entities.last
          #~ bore.pushpull(-@len_hub, false)

          #~ # insert hub and bore v.2
          #~ z0 = base_width / 2
          #~ z1 = @len_hub / 2
          #~ # draw hub and bore on each web face
          #~ [0,1].each{|n|
            #~ zHub = z0 - z1 if n == 0
            #~ zHub = z0 + z1 if n == 1
            #~ xHub = @rad_hub if @rad_hub < refRadius[n]
            #~ xHub = refRadius[n] if @rad_hub > refRadius[n]

            #~ hubCircle = entBase.add_circle([0,0,zHub], Z_AXIS, xHub, 32)
            #~ hubFace = entBase.add_face(hubCircle)
            #~ vect = [0, 0, -(hubFace.normal.z)]

            #~ ray = Sketchup.active_model.raytest [xHub, 0, zHub], vect, false
      #~ puts ray.inspect
            #~ zPush = ray[0].z
            #~ hubFace.pushpull(zPush, false)

            #~ circleBore = entBase.add_circle([0,0,zHub], Z_AXIS, @rad_bore, 32)
            #~ entities = entBase.to_a
            #~ bore = entities.last
            #~ bore.pushpull(-@len_hub, false)
          #~ }

        elsif @rad_bore > 0

          #~ # make a bore for the shaft, w/o a hub
          #~ z0 = web_depth
          #~ z1 = z0 + web_width
          #~ circleBore = entBase.add_circle([0,0,z0], Z_AXIS, @rad_bore, 32)
          #~ entities = entBase.to_a
          #~ bore = entities.last
          #~ bore.pushpull(-@len_hub, false)		

        end # draw hub and shaft bore

        # explode extrusion inside of parent group
      #	subGroup.transform! t1
      #	subGroup.explode

        #~ # collect points to for tooth end-surfaces
        #~ ptsReturn = []
        #~ ptsReturn[0] = profilePts.first
        #~ ptsReturn[1] = profilePts.last

        #~ ptsReturn   # return end-surface points

        grpBase.explode

      end  # spDrawBase

      #~ # IGNORE - only used by error handler during development
      #~ def LogToFile(obj)
        #~ currTime = Time.now().to_i
        #~ fileName = "log_file_" << currTime.to_s << ".txt"
        #~ Dir::chdir("C:/Users/Jason/Downloads/Sketchup Plugins")
        #~ logFile = File.new(fileName, "w")
        #~ if File.exists?(logFile)
          #~ logFile.syswrite(obj.to_s)
          #~ UI.messagebox("Log file created.\n" << Dir::pwd << fileName)		
        #~ else
          #~ UI.messagebox("failed to create dump file")
        #~ end
        #~ logFile = nil
      #~ end

      def spSoftSmooth(edges)
        begin # if arg contains a single edge, insert it into an array
        edges = [edges] if edges.typename == "Edge"
        rescue
        end # rescue

        edges.each{|e| e.soft = "true"; e.smooth = "true"}

      end # spSoftSmooth 

      # Points on involute curve parameterized by ang (in radians).
      def spGearInvolute(ang)
        x1 = @rad_base * Math.cos(ang)
        y1 = @rad_base * Math.sin(ang)
        s = ang * @rad_base
        x = x1 + s * Math.sin(ang)
        y = y1 - s * Math.cos(ang)

        # return x, y coordinates
        [x, y]

      end  # spGearInvolute

      def spFilletPoints(x0, y0)
        points = []
        #deg = 180 downto 90
        ang = 180.degrees
        4.times do
          x = x0 + @rad_fillet * Math.cos(ang)
          y = y0 + @rad_fillet * Math.sin(ang)
          points.push([x, y, 0])
          ang = ang - 180.degrees / 6.0
        end

        points

      end # spFilletPoints

      def spCopyGroup(objGroup, strSketchyShapeName, blnMergeObjects)
        # convert initial group to a component
        objComp = objGroup.to_component
        objCompDef = objComp.definition

        # create a new group to contain copies
        grpObjects = @mainGroup.entities.add_group

        # add instances to the group and rotate to their final position
        0.upto(@teeth-1){|n|
          t0 = Geom::Transformation.rotation [0,0,0], Z_AXIS, (360.degrees * n / @teeth)
          #~ t1 = Geom::Transformation.rotation [@rad,0,0], Y_AXIS, -@bevel_angle
          newComp = grpObjects.entities.add_instance(objCompDef, t0)
          newComp.set_attribute("SPOBJ", "shape", strSketchyShapeName)
          newComp.explode if blnMergeObjects
        }  # loop

        # erase original group/component
        objComp.erase!

        # if objects were merged, then apply shape name to the group
        grpObjects.set_attribute("SPOBJ", "shape", strSketchyShapeName) if blnMergeObjects

        grpObjects # return new group

      end # spCopyGroup

    end  # class SPGear

    # ----------------------------------------------------------------------------

    # User Interface Stuff - Run Once at Startup:
    if not defined?(@SPGears_loaded)
      menu = UI.menu("Plugins")
      item = menu.add_item('SPGears (MJG)') { SPGear.dialog }

      @SPGears_loaded = true
    end

  end # extension submodule
end # author namespace module
